/******************************************/
/*                                        */
/* Code Copyright (C) Terry Richards 1992 */
/*                                        */
/* FREE for personal use only. Commercial */
/* use forbidden without my express       */
/* written permission.                    */
/*                                        */
/******************************************/

#include <owl.h>
#include "bigedit.h"
#include <edit.h>
#include <checkbox.h>

#define max(a,b)(((a) >(b)) ?(a) :(b))
#define min(a,b)(((a) <(b)) ?(a) :(b))

TBFileWindow::TBFileWindow(PTWindowsObject AParent, LPSTR ATitle):TWindow(AParent, ATitle){

	IsNewFile  	= TRUE;
	Anchor		= NULL;
	HasFocus	= TRUE;
	DoubleClick	= FALSE;
	DidSearch	= FALSE;

	Attr.Style |= WS_VSCROLL | WS_HSCROLL;
	Scroller = new TBScroller(this,8,20,0,0);    // scroll parms reset later
	Scroller->AutoOrg = FALSE;                  //Our files are too big for AutoOrg to work right

    CaretWidth = GetSystemMetrics(SM_CXBORDER) * 2;

	Initialize();

	SetScroller();

}; 

void TBFileWindow::Initialize(){

//Set all the stuff that relates to a new file

	ButtonDown  = FALSE;
	IsDirty     = FALSE;
	IsKB        = FALSE;

	Size		= 0;
	CaretPos	= 0;
	HiStart		= 0;
    HiEnd		= 0;

	Scroller->ScrollTo(0,0);

    DeleteBuffers();
	Anchor = new TBFileBuffer(this);

};

TBFileWindow::~TBFileWindow(){

	DeleteBuffers();

}

void TBFileWindow::Paint(HDC DC, PAINTSTRUCT&){

//COLORREF	Color;
HBRUSH		Brush,OldBrush;
HPEN		Pen,OldPen;
HBITMAP		BitMap;
int 		AllBitsOn[] = {0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF};
POINT		CursorPos;
RECT		WindowSize;

	//This code creates a really cool brush that draws in
	//the current text background color. Great for hilite
	//text where we switch the background & foreground
    //colors.

	BitMap = CreateBitmap(8,8,1,1,AllBitsOn);
	Brush = CreatePatternBrush(BitMap);
	(HGDIOBJ)OldBrush = SelectObject(DC,(HGDIOBJ)Brush);

	//Use a null pen as we don't want the rectangles to have borders.

	Pen = CreatePen(PS_NULL,1,RGB(0,0,0));
	(HGDIOBJ)OldPen = SelectObject(DC,(HGDIOBJ)Pen);

	SetTextAlign(DC,TA_UPDATECP);   /* updates current position */

	//If we are dragging then the mouse could be stationary above or
	//below the window and WMMouseMove() won't be getting any messages.
	//If the cursor is off the screen we set HiEnd to the extreme
	//position (0 or Size) which will cause the display to be correct.
	//At the end of scrolling it will all get sorted out.

	if (ButtonDown){

		GetCursorPos(&CursorPos);
		ScreenToClient(HWindow,&CursorPos);

		if (CursorPos.y < 0)
			HiEnd = 0;

		else{

			GetClientRect(HWindow,&WindowSize);
			if (CursorPos.y > WindowSize.bottom)
				HiEnd = Size;

		}

	}

	Anchor->Paint(DC,Scroller,HiStart,HiEnd);
	SetCaret();

	//Tidy up

	SelectObject(DC,OldBrush);
    DeleteObject(Brush);
	SelectObject(DC,OldPen);
	DeleteObject(Pen);
	DeleteObject(BitMap);

}

void TBFileWindow::SetScroller(){

long    XRange=0,YRange=0;

	XRange = Anchor->XRange() / Scroller->XUnit;
	YRange = Anchor->YRange() / Scroller->YUnit;
	
	Scroller->SetRange(XRange,YRange);

}

void TBFileWindow::SetCaret(){

long    X,Y,XOffset=0,YOffset=0;
int     H;
HDC     hdc;

	if (!HasFocus || ButtonDown)
    	return;

	DestroyCaret();

	hdc = GetDC(HWindow);
	Anchor->BufferToClient(CaretPos,hdc,X,Y);
	ReleaseDC(HWindow,hdc);

	H = Anchor->Height(CaretPos);

	//On Keyboard input we allow automatic scrolling
	//We change X & Y independantly so we can scroll 
	//"Just Enough"
		
	//Did we move off the left edge?
		
	XOffset = X - (Scroller->XPos * Scroller->XUnit);
		
	if (XOffset > 0){
		
		//No - Did we move off the right edge?
			
		XOffset -= Scroller->XPage * Scroller->XUnit;
			
		if (XOffset < 0)
			
			//No - we don't need an X scroll
				
			XOffset = 0;
				
	}

	//Same logic for top/bottom
		
	YOffset = Y - (Scroller->YPos * Scroller->YUnit);
		
	if (YOffset > 0){
		
		YOffset -= Scroller->YPage * Scroller->YUnit;
			
		if (YOffset < 0)
			YOffset = 0;
				
	}
	
	if ((XOffset || YOffset) && IsKB){

		//This redraw was caused by keyboard cursor keys - Autoscroll.

        //First make sure it will scroll

		if (XOffset < 0)
			XOffset = (XOffset/Scroller->XUnit) - 1;
		else if (XOffset > 0)
			XOffset = (XOffset/Scroller->XUnit) + 1;

		if (YOffset < 0)
			YOffset = (YOffset/Scroller->YUnit) - 1;
		else if (YOffset > 0)
			YOffset = (YOffset/Scroller->YUnit) + 1;

		//Then scroll

		IsKB = FALSE; //Scroller calls Paint & we don't want a loop!
		Scroller->ScrollBy(XOffset,YOffset);

    }

	IsKB = FALSE;

	CreateCaret(HWindow,0,CaretWidth,H);
	SetCaretPos(X-(Scroller->XPos * Scroller->XUnit),Y-(Scroller->YPos * Scroller->YUnit));
	ShowCaret(HWindow);

}

BOOL TBFileWindow::CanClose(){

	if (!IsDirty)
		return TRUE;

	return IDNO == MessageBox(HWindow,"Do you want to save?","File has changed",MB_YESNO | MB_ICONQUESTION);

}

void TBFileWindow::WMLButtonDown(RTMessage Msg){

long X,Y;

	InvalidateRect(HWindow,NULL,FALSE);

	if (!ButtonDown){

		ButtonDown = TRUE;
		SetCapture(HWindow);
		DragDC = GetDC(HWindow);

	}

	X = Msg.LP.Lo + (Scroller->XPos * Scroller->XUnit);
	Y = Msg.LP.Hi + (Scroller->YPos * Scroller->YUnit);

	if (Msg.WParam & MK_SHIFT) //Shift-Click

		HiEnd = Anchor->ClientToBuffer(X,Y,DragDC);

	else{

		//Move the caret to the mouse position. No need to
		//redraw as the Paint function will fix it.

		CaretPos = Anchor->ClientToBuffer(X,Y,DragDC);

		HiStart = CaretPos;
		HiEnd	= CaretPos;

    }

	HideCaret(HWindow);

	//Pass the event to the underlying TWindow
	//It does useful things with it. 

	TWindow::WMLButtonDown(Msg);

}

void TBFileWindow::WMMouseMove(RTMessage Msg){

	if (ButtonDown && !DoubleClick){

		HiEnd = Anchor->ClientToBuffer(Msg.LP.Lo + (Scroller->XPos * Scroller->XUnit),Msg.LP.Hi + (Scroller->YPos * Scroller->YUnit),DragDC);
		InvalidateRect(HWindow,NULL,FALSE);

		//Force a repaint right now or the WM_PAINT
		//message will be lost in a stream of rodent
		//motion messages. 

		UpdateWindow(HWindow);

	}

}

void TBFileWindow::WMLButtonUp(RTMessage Msg){

long	Temp;

	if (ButtonDown){

		if (!DoubleClick)
			HiEnd = Anchor->ClientToBuffer((signed)Msg.LP.Lo + (Scroller->XPos * Scroller->XUnit),
											(signed)Msg.LP.Hi + (Scroller->YPos * Scroller->YUnit),DragDC);

		ButtonDown	= FALSE;
        DoubleClick	= FALSE;

		if (HiEnd < HiStart){

			Temp		= HiEnd;
			HiEnd		= HiStart;
			HiStart		= Temp;
			CaretPos	= HiStart;

        }

		InvalidateRect(HWindow,NULL,FALSE);
		ReleaseCapture();
		ReleaseDC(HWindow,DragDC);

	}

}

void TBFileWindow::WMLButtonDblClk(RTMessage Msg){

long X,Y;

	InvalidateRect(HWindow,NULL,FALSE);

    DoubleClick = TRUE;

	if (!ButtonDown){

		ButtonDown = TRUE;
		SetCapture(HWindow);
		DragDC = GetDC(HWindow);

	}

	X = Msg.LP.Lo + (Scroller->XPos * Scroller->XUnit);
	Y = Msg.LP.Hi + (Scroller->YPos * Scroller->YUnit);

	CaretPos = Anchor->ClientToBuffer(X,Y,DragDC);

	HiStart		= Anchor->PrevBreak(CaretPos);
	HiEnd		= Anchor->NextBreak(CaretPos);
    CaretPos	= HiStart;

}

void TBFileWindow::WMKeyDown(RTMessage Msg){

long	X,Y;
HDC		hdc;

	IsKB = FALSE;

	switch (Msg.WParam) {

		case VK_LEFT:

			if (CaretPos > 0){

				CaretPos--;
				IsKB = TRUE;

			}

			break;

		case VK_RIGHT:

			if (CaretPos < Size){

				CaretPos++;
				IsKB = TRUE;

			}

			break;

		case VK_DOWN:

			hdc = GetDC(HWindow);

			Anchor->BufferToClient(CaretPos,hdc,X,Y);

            Y += Anchor->Height(CaretPos);
			CaretPos = Anchor->ClientToBuffer(X,Y,hdc);

			IsKB = TRUE;
            ReleaseDC(HWindow,hdc);

			break;

		case VK_NEXT: //Page Down

			hdc = GetDC(HWindow);

			Anchor->BufferToClient(CaretPos,hdc,X,Y);

			Y += Scroller->YPage * Scroller->YUnit;
			CaretPos = Anchor->ClientToBuffer(X,Y,hdc);

			IsKB = TRUE;
            ReleaseDC(HWindow,hdc);

			break;

		case VK_UP:

			hdc = GetDC(HWindow);

			Anchor->BufferToClient(CaretPos,hdc,X,Y);

			if (Y != 0){

	            Y -= Anchor->Height(CaretPos);
				CaretPos = Anchor->ClientToBuffer(X,Y,hdc);
				IsKB = TRUE;

			}

            ReleaseDC(HWindow,hdc);

			break;

		case VK_PRIOR:	//Page up

			hdc = GetDC(HWindow);

			Anchor->BufferToClient(CaretPos,hdc,X,Y);

			Y -= Scroller->YPage * Scroller->YUnit;

			if (Y < 0)
				Y = 0;

			CaretPos = Anchor->ClientToBuffer(X,Y,hdc);

			IsKB = TRUE;
            ReleaseDC(HWindow,hdc);

			break;

		case VK_HOME:

			CaretPos = Anchor->PrevCR(CaretPos);
			IsKB = TRUE;

			break;

		case VK_END:

			CaretPos = Anchor->NextCR(CaretPos);
			IsKB = TRUE;

			break;

		case VK_BACK:

			//If there is hilighted text delete it all

			if (HiStart < HiEnd){

				Anchor->DeleteAt(HiStart,HiEnd - HiStart);
				Size -= HiEnd - HiStart;

				SetScroller(); //Dataset Dimensions changed
				IsKB    = TRUE;
				IsDirty = TRUE;

            }

			//Otherwise just delete one character

			if (CaretPos > 0){

				CaretPos--;
				Anchor->DeleteAt(CaretPos);
				Size--;
				SetScroller(); //Dataset Dimensions changed
				IsKB    = TRUE;
				IsDirty = TRUE;

            }

			break;

		case VK_DELETE:

        	//If there is hilighted text delete it all

			if (HiStart < HiEnd){

				Anchor->DeleteAt(HiStart,HiEnd - HiStart);
				Size -= HiEnd - HiStart;

				SetScroller(); //Dataset Dimensions changed
				IsKB    = TRUE;
				IsDirty = TRUE;

            }

			//Otherwise just delete one character

			else if (CaretPos < Size){

				Anchor->DeleteAt(CaretPos);
				Size--;
				SetScroller(); //Dataset Dimensions changed
				IsKB    = TRUE;
				IsDirty = TRUE;

            }

			break;

	}

	if (IsKB){

    	HiEnd = HiStart; //remove the highlight
		InvalidateRect(HWindow,NULL,FALSE);

    }

	DefWndProc(Msg);

}

void TBFileWindow::WMChar(RTMessage Msg){

long	X,Y;
HDC		hdc;

	IsKB = FALSE;

	if (Msg.WParam > VK_ESCAPE || Msg.WParam == VK_TAB || Msg.WParam == VK_RETURN) {

		if (Msg.WParam == VK_RETURN)
			Msg.WParam = '\n';

		//If there is hilighted text overwrite it

		if (HiStart < HiEnd){

			Anchor->DeleteAt(HiStart,HiEnd - HiStart);
			Size -= HiEnd - HiStart;

        }

		Anchor->InsertAt(CaretPos,Msg.WParam);
		Size++;
		CaretPos++;
		SetScroller(); //Dataset Dimensions changed
		IsKB    = TRUE;
		IsDirty = TRUE;

	}

	if (IsKB){

    	HiEnd = HiStart; //remove the highlight
		InvalidateRect(HWindow,NULL,FALSE);

    }

	DefWndProc(Msg);

}

void TBFileWindow::WMKillFocus(RTMessage){

	HasFocus = FALSE;
	DestroyCaret();

}

void TBFileWindow::WMSetFocus(RTMessage){

	HasFocus = TRUE;
	SetCaret();

}

void TBFileWindow::CMFileNew(RTMessage){

	if (!CanClose())
		return;

	InvalidateRect(HWindow,NULL,TRUE);
	IsNewFile = TRUE;
	Initialize();
	SetScroller();

}

void TBFileWindow::CMFileOpen(RTMessage){

	if (!CanClose())
		return;

	if(GetApplication()->ExecDialog(new TFileDialog(this,SD_FILEOPEN,strcpy(FileName,"*.*"))) == IDOK )
		OpenFile();

}

void TBFileWindow::CMFileSave(RTMessage){

	if (IsNewFile)
		SaveFileAs();
	else
		SaveFile();

}

void TBFileWindow::CMEditCut(RTMessage){

	if (!EditCopy())
		return;

	Anchor->DeleteAt(HiStart,HiEnd-HiStart);
	Size -= (HiEnd-HiStart);
	HiEnd = HiStart;
	CaretPos = HiStart;
	IsDirty = TRUE;
	SetScroller();
	InvalidateRect(HWindow,NULL,FALSE);

}

void TBFileWindow::CMEditCopy(RTMessage){

	EditCopy();

}

void TBFileWindow::CMEditPaste(RTMessage){

HANDLE		hMemory;
char FAR	*String;
char 		*NewString;
long	 	i,j;

	if (!IsClipboardFormatAvailable(CF_TEXT))
		return; //Nothing to paste

    //First open the clipboard which gives us ownership

	if (!OpenClipboard(HWindow)){

		MessageBox(HWindow,"Can't lock clipboard","Clipboard Error",MB_ICONEXCLAMATION | MB_OK);
		return;

    }

	//Get the clipboard data and convert to a pointer

	hMemory = GetClipboardData(CF_TEXT);

	if (!hMemory){

    	CloseClipboard();
		MessageBox(HWindow,"Can't get clipboard data","Clipboard Error",MB_ICONEXCLAMATION | MB_OK);
		return;

    }

	String = (char FAR *)GlobalLock(hMemory);

	//We have lines terminated with \n and the clipboard
	//uses \r\n so we have to compress them.

	//Allocate enough memory to copy the string to

	NewString = new char[strlen(String)+1];

	if (!NewString){

		MessageBox(HWindow,"Unable to allocate memory block 2","Memory Error",MB_OK | MB_ICONEXCLAMATION);
		return;

	}

	//Now copy the old data compressing as we go.

	for (i=0,j=0;String[i];i++)

		if (String[i] != '\r'){

           	NewString[j] = String[i];
			j++;

		}

	NewString[j] = 0;

	//return control of the memory block to the clipboard -
    //we are done with it.

	GlobalUnlock(hMemory);
    CloseClipboard();

	//Pasting overwrites any hilighted text

	if (HiStart != HiEnd){

		Anchor->DeleteAt(HiStart,HiEnd-HiStart);
		Size -= (HiEnd - HiStart);
		HiEnd = HiStart;
        CaretPos = HiStart;

	}

	//Now insert the new data

	Anchor->InsertAt(CaretPos,NewString);
	Size += strlen(NewString);
    IsDirty = TRUE;
	InvalidateRect(HWindow,NULL,FALSE);
	SetScroller();

	//Free the copy memory

    delete NewString;

}

void TBFileWindow::CMEditDelete(RTMessage){

	if (HiStart == HiEnd)
		return; //Nothing to delete

	Anchor->DeleteAt(HiStart,HiEnd-HiStart);
	Size -= (HiEnd - HiStart);
	HiEnd = HiStart;
	CaretPos = HiStart;
	IsDirty = TRUE;
	InvalidateRect(HWindow,NULL,FALSE);
	SetScroller();

}

void TBFileWindow::CMEditClear(RTMessage){

	//The user wants to clear the editor. We take this
	//chance to clean up the buffer chain by deleting it
	//completely and starting again.

	InvalidateRect(HWindow,NULL,TRUE);
	Initialize();
	SetScroller();

}

void TBFileWindow::CMEditUndo(RTMessage){

	MessageBox(HWindow,"Undo not implemented yet","Sorry",MB_ICONEXCLAMATION | MB_OK);

}

void TBFileWindow::CMEditFind(RTMessage){

char	Work[sizeof(SearchStruct.SearchText)+26];

	if (GetModule()->ExecDialog(new TBSearchDialog(this,SD_SEARCH,&SearchStruct,GetModule())) == IDOK ){

        DidSearch = TRUE;

		SearchStruct.Found		= FALSE;
		SearchStruct.StartPos	= CaretPos;

		//NOTE: StartPos is not preserved by Search()
		//DO NOT assume anything about its value after
        //Search() returns.

        Anchor->Search(SearchStruct);

		if (SearchStruct.Found){

			HiStart		= SearchStruct.FoundAt;
            HiEnd		= HiStart + strlen(SearchStruct.SearchText);
            CaretPos	= HiEnd;
			IsKB 		= TRUE; //Forces auto-scrolling

		}

		else{

			HiEnd = HiStart;
			strcpy(Work,"String <<");
			strcat(Work,SearchStruct.SearchText);
            strcat(Work,">> Was Not Found");
			MessageBox(HWindow,Work,"Search Failed",MB_ICONEXCLAMATION | MB_OK);

        }

		//Even if the search failed the highlight could have changed

		InvalidateRect(HWindow,NULL,FALSE);
		UpdateWindow(HWindow);

	}

}

void TBFileWindow::CMEditFindNext(RTMessage){

char	Work[sizeof(SearchStruct.SearchText)+26];

	//If we didn't previously do a search then
	//we can't search again.

	if (!DidSearch)
    	return;

	SearchStruct.Found		= FALSE;
	SearchStruct.StartPos	= CaretPos;

	//NOTE: StartPos is not preserved by Search()
	//DO NOT assume anything about its value after
	//Search() returns.

	Anchor->Search(SearchStruct);

	if (SearchStruct.Found){

		HiStart		= SearchStruct.FoundAt;
		HiEnd		= HiStart + strlen(SearchStruct.SearchText);
		CaretPos	= HiEnd;
		IsKB 		= TRUE; //Forces auto-scrolling

	}

	else{

    	HiEnd = HiStart;
		strcpy(Work,"String <<");
		strcat(Work,SearchStruct.SearchText);
		strcat(Work,">> Was Not Found");
		MessageBox(HWindow,Work,"Search Failed",MB_ICONEXCLAMATION | MB_OK);

    }

	//Even if the search failed the highlight could have changed

	InvalidateRect(HWindow,NULL,FALSE);
    UpdateWindow(HWindow);

}

void TBFileWindow::CMEditReplace(RTMessage){

char	Work[sizeof(SearchStruct.SearchText)+26];
int		rc=IDYES;

	if (GetModule()->ExecDialog(new TBSearchDialog(this,SD_REPLACE,&SearchStruct,GetModule())) == IDOK ){

        DidSearch = TRUE;

		SearchStruct.Found		= FALSE;
		SearchStruct.StartPos	= CaretPos;

		//NOTE: StartPos is not preserved by Search()
		//DO NOT assume anything about its value after
        //Search() returns.

        Anchor->Search(SearchStruct);

		if (!SearchStruct.Found){

			HiEnd = HiStart;
			strcpy(Work,"String <<");
			strcat(Work,SearchStruct.SearchText);
            strcat(Work,">> Was Not Found");
			MessageBox(HWindow,Work,"Search Failed",MB_ICONEXCLAMATION | MB_OK);
			InvalidateRect(HWindow,NULL,FALSE);

            return;

        }

		while (SearchStruct.Found){

			HiStart		= SearchStruct.FoundAt;
            HiEnd		= HiStart + strlen(SearchStruct.SearchText);
            CaretPos	= HiEnd;
			IsKB 		= TRUE; //Forces auto-scrolling

			if (SearchStruct.PromptOnReplace){

				InvalidateRect(HWindow,NULL,FALSE);
				UpdateWindow(HWindow);
				rc = MessageBox(HWindow,"Replace this occurance?","Search/Replace",MB_ICONQUESTION | MB_YESNOCANCEL);

			}

			if (rc == IDCANCEL)
				break;

			else if (rc == IDYES){

				Anchor->DeleteAt(HiStart,strlen(SearchStruct.SearchText));
				Anchor->InsertAt(HiStart,SearchStruct.ReplaceText);
				CaretPos = HiStart + strlen(SearchStruct.ReplaceText);
				HiEnd = HiStart;
				IsDirty = TRUE;

			}

			if (!SearchStruct.ReplaceAll)
            	break;

			SearchStruct.Found		= FALSE;
			SearchStruct.StartPos	= CaretPos;

			Anchor->Search(SearchStruct);

		}

		//Even if the search failed the highlight could have changed

		InvalidateRect(HWindow,NULL,FALSE);
		SetScroller();
		UpdateWindow(HWindow);

	}

}

void TBFileWindow::GetWindowClass(WNDCLASS& AClass){

	TWindow::GetWindowClass(AClass);

	AClass.hCursor = LoadCursor(NULL,IDC_IBEAM);
	AClass.style |= CS_DBLCLKS;

}


void TBFileWindow::SaveFileAs(){

	if (IsNewFile)
		strcpy(FileName, "");

	if (GetApplication()->ExecDialog(new TFileDialog(this,SD_FILESAVE,strcpy(FileName,"*.*"))) == IDOK)
		SaveFile();

}

void TBFileWindow::CMFileSaveAs(RTMessage){

	SaveFileAs();

}

void TBFileWindow::SaveFile(){

ofstream		os(FileName);
PTBFileBuffer	Cursor;

	Cursor = Anchor;

	do {

		Cursor->UnLoad(os);

		if (os.bad()){

			MessageBox(HWindow,"Unable to write file","File Error",MB_OK | MB_ICONEXCLAMATION);
			return;

		}

		Cursor = Cursor->Next;

	} while (Cursor);

	os.close();

	IsNewFile = IsDirty = FALSE;

}

void TBFileWindow::OpenFile(){

ifstream        is(FileName);
PTBFileBuffer	Last;
HCURSOR         hCursor;
HDC             hdc;

	if (is.bad()){

		MessageBox(HWindow,"Unable to open file","File Error",MB_OK | MB_ICONEXCLAMATION);
		return;

	}

	Initialize();
	Last = Anchor;

	hCursor = SetCursor(LoadCursor(NULL,IDC_WAIT));

	while (TRUE){

		if (!Last){

			MessageBox(HWindow,"Could not allocate Buffer","Memory Error",MB_OK | MB_ICONEXCLAMATION);
			DeleteBuffers();
			SetScroller();
			SetCursor(hCursor);
			return;

		}

		Size += Last->Load(is);

		if (is.eof())
			break;

		if (is.bad()){

			MessageBox(HWindow,"Read Error","File Error",MB_OK | MB_ICONEXCLAMATION);
			DeleteBuffers();
			SetScroller();
			SetCursor(hCursor);
			return;

		}

		Last = new TBFileBuffer(this,Last);

	}

	is.close();

	IsNewFile = FALSE;
	SetScroller();
	InvalidateRect(HWindow,NULL,TRUE);
	SetCursor(hCursor);

}

BOOL TBFileWindow::EditCopy(){

HGLOBAL		hGlobalMemory,hNewMemory;
char FAR*	lpString;
char FAR*	lpNewString;
long		cCR=0,i,j;

	if (HiStart == HiEnd) //Nothing to copy
		return FALSE;

	//Huge arrays aren't supported so we limit copying
	//and cutting to one data segment. The only way I can
	//see around this is to define a private clipboard
	//format. Then we wouldn't be able to paste into
	//regular small editors. It's a problem!
	//Flat 32 bit memory will cure it.

	if (HiEnd - HiStart >= MAXINT){

		MessageBox(HWindow,"Too big for the Clipboard","Memory Error",MB_OK | MB_ICONEXCLAMATION);
		return FALSE;

	}

	//Allocate enough memory to copy the string to

	hGlobalMemory = GlobalAlloc(GHND,HiEnd-HiStart+1); //NULL terminated

	if (!hGlobalMemory){

		MessageBox(HWindow,"Unable to allocate memory block","Memory Error",MB_OK | MB_ICONEXCLAMATION);
		return FALSE;

	}

	//Nail it down so we can write to it without it moving

	lpString = (char FAR *) GlobalLock(hGlobalMemory);

	//Copy the needed part of the buffer

	Anchor->CopyChar(lpString,HiStart,HiEnd-HiStart);

	//We have lines terminated with \n and the clipboard
	//needs \r\n so we have to expand them. First find out
	//how many there are.

	for (i=0;i < HiEnd-HiStart;i++)
		if (lpString[i] == '\n')
			cCR++;

	//If there are any then we have to convert.

	if (cCR){

		//Allocate enough memory to copy the string to

		hNewMemory = GlobalAlloc(GHND,HiEnd-HiStart+cCR+1); //NULL terminated

		if (!hNewMemory){

			MessageBox(HWindow,"Unable to allocate memory block 2","Memory Error",MB_OK | MB_ICONEXCLAMATION);
			return FALSE;

		}

	    //Nail it down so we can write to it without it moving

		lpNewString = (char FAR *) GlobalLock(hNewMemory);

		//Now copy the old data expanding as we go.

		for (i=0,j=0;i < HiEnd - HiStart;i++,j++){

			if (lpString[i] == '\n'){

            	lpNewString[j] = '\r';
				j++;

			}

			lpNewString[j] = lpString[i];

		}

		//Very carefully! We are going to delete the old memory block
		//and reset the handles so the new block looks like the old one.
		//This includes the fact that the old block is locked at this point.

		GlobalUnlock(hGlobalMemory);
		GlobalFree(hGlobalMemory);
		hGlobalMemory = hNewMemory;

    }

	//We are going to give this block of memory to the clipboard so
	//we *MUST* unlock it. We *MUST NOT* free it though - it isn't ours
	//any more.

	GlobalUnlock(hGlobalMemory);

	OpenClipboard(HWindow);
	EmptyClipboard();
	SetClipboardData(CF_TEXT,hGlobalMemory);
	CloseClipboard();

    return TRUE;

}

void TBFileWindow::DeleteBuffers(){

PTBFileBuffer	Cursor;
PTBFileBuffer	Next;

	//We have redundant calls to this function so we
	//make it safe to be called multiple times. That way
	//we never have orphaned or duplicate buffers. 

	if (!Anchor)
		return;

	Cursor = Anchor;

	do {

		Next = Cursor->Next;
		delete Cursor;
		Cursor = Next;

	} while (Cursor);

	Anchor = NULL;

}

TBFileBuffer::TBFileBuffer(PTWindowsObject AParent,PTBFileBuffer APrev){

TEXTMETRIC  tm;
HDC         hdc;

	Parent = AParent;

	//Set up forward & backward pointer chain

	if (APrev)
		APrev->Next = this;

	Prev = APrev;
	Next = NULL;

	//There's nothing in this buffer yet

	Size            = 0;
	cCR             = 0;
	LongestLine     = 0;

	//Set some text display defaults
	//A defined class will probably do 
	//something more interesting here

	hdc = GetDC(Parent->HWindow);
	GetTextMetrics(hdc,&tm);
	ReleaseDC(Parent->HWindow,hdc);

	CharAveWidth    = tm.tmAveCharWidth;
	LineHeight      = tm.tmHeight + tm.tmExternalLeading;
	TabSize         = 4 * CharAveWidth;
	BGColor 		= GetSysColor(COLOR_BACKGROUND);
	TextColor		= !BGColor;

}

long TBFileBuffer::Load(ifstream& is){

	//Read buffer from file

	is.read(Buffer,BUFFSIZE);

	//Save the number of bytes read

	if (is.eof())
		Size = is.gcount();
	else
		Size = BUFFSIZE;

	//Calculate some statistics so we can calculate
	//metrics faster when we need to.

	SetStats();

	return Size;

}

void TBFileBuffer::SetStats(){

long	i;
int		cChar=0;
BOOL    First=TRUE;

	cCR = 0;

	for (i=0;i<Size;i++){

		cChar++;

		if (Buffer[i] == '\n'){

			cCR++;

			if (First){

				First = FALSE;

				if (Prev)
					cChar += Prev->LastChars();

			}

			if (cChar > LongestLine)
				LongestLine = cChar;

			cChar = 0;

		}

	}

}

long TBFileBuffer::LastChars(){

long    i; 	//should be an unsigned int but that
			//would make the for loop endless

long    Count;

	//Parse the buffer backwards looking
	//for the end of the previous line.

    //Be very careful to maintain the sign.

	i = Size;
	i--;

	for (;i>=0 && Buffer[i] != '\n';i--);

	Count = Size;
	Count = Count - i - 1;

	if (i<0 && Prev)
		Count += Prev->LastChars();

	return Count;

}

void TBFileBuffer::UnLoad(ofstream& os){

	//Write buffer to file

	os.write(Buffer,(int)Size);

}

void TBFileBuffer::Paint(HDC DC,TScroller *Scroller,long HiStart,long HiEnd,long CharsSoFar,long YPos,BOOL First){

char    *String,*Cursor,*End,*HStart,*HEnd;
DWORD   dw;
RECT	Rect;
int		Top,Left,Bottom,Right;
int     X;
BOOL	HasHiLite = FALSE;
long	Temp;

	//if we are past the visible area we are done
	//Note that we check one extra line just to be sure

	if (YPos > ((long)(Scroller->YPos) + (long)((Scroller->YPage)+1)) * (long)(Scroller->YUnit))
		return;

	//If the bottom of this buffer is still above the screen there
	//is no reason to paint it. Note that there may be a partial line
	//with no CR in this buffer so we test one extra line just to
	//be certain. 

	if ((YPos + ((cCR + 1) * (long)LineHeight)) >= (Scroller->YPos * Scroller->YUnit)){

		//At least part of this buffer is visible - paint it.

		if (HiEnd < HiStart){

			Temp		= HiEnd;
			HiEnd		= HiStart;
			HiStart		= Temp;

        }

		String  = Cursor = Buffer;
		End     = Buffer + Size;
		HStart	= Buffer + HiStart;
        HEnd	= Buffer + HiEnd;

		//We set the text display attributes in every Buffer even though they
		//are the same in this class. Derived classes will probably change
        //them.

		if (HiStart < HiEnd)
			HasHiLite = TRUE;

		if (HiStart <= 0 && HiEnd > 0 && HasHiLite){

			SetTextColor(DC,BGColor);
			SetBkColor(DC,TextColor);

        }

		else{

			SetTextColor(DC,TextColor);
			SetBkColor(DC,BGColor);

        }

		if (First){

			MoveTo(DC,-(Scroller->XPos * Scroller->XUnit),YPos - (Scroller->YPos * Scroller->YUnit));
			First = FALSE;

		}

		while (TRUE){

			//Spin Cursor Forward to next CR, next Tab, next color change or end 

			while (*Cursor != '\n' && *Cursor != '\t' && Cursor != End && ((Cursor != HStart && Cursor != HEnd) || !HasHiLite))
				Cursor ++; 

			TextOut(DC,0,0,String,(int)(Cursor-String));

			if (Cursor == End)
				break;

			if (Cursor == HStart){

				SetTextColor(DC,BGColor);
				SetBkColor(DC,TextColor);

				if(*Cursor != '\n' && *Cursor != '\t' && Cursor != End)
					TextOut(DC,0,0,Cursor,1);

			}

			if (Cursor == HEnd){

				SetBkColor(DC,BGColor);
				SetTextColor(DC,TextColor);

				if(*Cursor != '\n' && *Cursor != '\t' && Cursor != End)
					TextOut(DC,0,0,Cursor,1);

			}

			if (*Cursor == '\n'){

				//For a line break move to the start
				// of the next line

				dw = GetCurrentPosition(DC);

				//Obliterate any old caret or text - this saves erasing
				//the whole window which causes an annoying flicker.
				GetClientRect(Parent->HWindow,&Rect);
				Left	= LOWORD(dw);
				Top		= HIWORD(dw);
				Right	= Rect.right + 1;
				Bottom	= Top + LineHeight + 1;

				Rectangle(DC,Left,Top,Right,Bottom);

				MoveTo(DC,-(Scroller->XPos * Scroller->XUnit),HIWORD(dw)+LineHeight);

			}

			if (*Cursor == '\t'){

				//For a tab move to the next tab position
				//I can't seem to make TabbedTextOut work

				dw = GetCurrentPosition(DC);
				X = LOWORD(dw) + (Scroller->XPos * Scroller->XUnit);
				X /= TabSize;
				X++;
				X *= TabSize;

				//Obliterate any old caret or text - this saves erasing
				//the whole window which causes an annoying flicker.

				Left	= LOWORD(dw);
				Top		= HIWORD(dw);
				Right	= Left + TabSize + 1;
				Bottom	= Top + LineHeight + 1;

                Rectangle(DC,Left,Top,Right,Bottom);

				MoveTo(DC,X - (Scroller->XPos * Scroller->XUnit),HIWORD(dw));

			}

			String = ++Cursor;

		}

	}

	if (Next)
		Next->Paint(DC,Scroller,HiStart - Size,HiEnd - Size,CharsSoFar - Size,YPos + (cCR * LineHeight),First);

	else{

		//If we get here there could be a partial screen drawn
		//Erase the rest of the screen

		SetTextColor(DC,TextColor);
		SetBkColor(DC,BGColor);

		//First do the rest of the current line

		dw = GetCurrentPosition(DC);
		GetClientRect(Parent->HWindow,&Rect);

		Left	= LOWORD(dw);
		Top		= HIWORD(dw);
		Right	= Rect.right + 1;
		Bottom	= Top + LineHeight + 1;    //Make sure we get them all

		Rectangle(DC,Left,Top,Right,Bottom);

		//Now do the bottom of the screen

		Left	= 0;
		Top		= Bottom - 1;
		Bottom	= Rect.bottom + 1;

		Rectangle(DC,Left,Top,Right,Bottom);

	}

}

long TBFileBuffer::XRange(long XSoFar){

long    Width;

	//I don't know how to do this efficiently AND accurately
	//The "perfect" way would be to GetTextExtent of every
	//line in the dataset and return the biggest. Another way
	//would be to use the average char width and fudge it up
	//a bit to make sure we have enough range. Borland's TCW
	//Editor just seems to "pick a number" and truncate any
	//lines that won't fit inside this window size. They
	//don't know either!

	//The second method seems the best compromise. TScrollers
	//give you an extra page width which should take care of
	//the fudge part.

	Width = (long)LongestLine * (long)CharAveWidth;

	if (Width > XSoFar)
		XSoFar = Width;

	if (Next)
		return Next->XRange(XSoFar); 

	return XSoFar;

}

long TBFileBuffer::YRange(long YSoFar){

long Lines;

	if (Next)
		return Next->YRange(YSoFar + (cCR * LineHeight));

	//If we get here then this is the last buffer in the chain
	//If the last character is not a <CR> then we are one line short

	Lines = cCR;

	if (Size > 0 && Buffer[Size-1] != '\n')
		Lines++;

	return YSoFar + (Lines * LineHeight);

}

void TBFileBuffer::BufferToClient(long BufferPos,HDC hdc,long& X,long& Y,long YSoFar,long XOffset){

long	i;
int		numCR=0;
long    LineLength=0;
DWORD   dw;

	if (BufferPos > Size){

		if (Next){

			//Calculate how much this buffer offsets
			//the co-ords. Y is easy - just the number
			//of lines times the line height.

			YSoFar += cCR * LineHeight;

			//X is tougher. If there is a part line
			//then it is the width of that line. If
			//the whole buffer is a part line then
			//it is added to the prior offset. If the
			//Buffer breaks at a kerning point then this
			//will cause a small error. A "better" approach
			//would be to actually string the entire line
			//together and measure it. But ... How long is
			//it? It could exceed a data segment length.

			//Good luck if you try this :-)

			//NOTE: The paint routine has the same error
			//which means that the co-ords will always match
			//the text and current windows fonts don't kern
			//anyway. 

			if (cCR){

				i = Size;
                i--;
				for (;i>=0 && Buffer[i] != '\n';i--);

			}

			else
				i = -1;

			i += 1;

			dw = GetTabbedTextExtent(hdc,Buffer+i,Size-i,1,&TabSize);

			if (cCR)
				XOffset = 0;

			XOffset += LOWORD(dw);

			Next->BufferToClient(BufferPos-Size,hdc,X,Y,YSoFar,XOffset);

			return;

		}

		else
			BufferPos = Size + 1; //Just after the last char.

	}

	//If we get here then we are in the right buffer

	//Parse through the buffer counting the number of CRs
	//before the caret position. Keep track of the part
	//line leading up to the buffer position too.

	for (i=0;i<BufferPos;i++){

		LineLength++;

		if (Buffer[i] == '\n'){

			numCR++;
			LineLength = 0;

		}

	}

	//The X offset is only meaningful if the co-ord is in
	//a partial line at the start of the buffer.
	//If there was a carriage return before it then blow
	//away the offset

	if (numCR)
		XOffset = 0; 

	dw = GetTabbedTextExtent(hdc,Buffer + BufferPos - LineLength,LineLength,1,&TabSize);

	Y = YSoFar + (numCR * LineHeight);
	X = XOffset + LOWORD(dw);

	//finally! we have the caret position in client coords.
	//The joys of proportional fonts are almost limitless.

}

long TBFileBuffer::ClientToBuffer(long X,long Y,HDC DC,long CharsSoFar,long XOffset){

long    ThisY;
long    i;
int		NeededCR,LastX,ThisX;
DWORD   dw;
char    *String;

	//Do the easy case first. If this position is
	//past this buffer decrement the Y, set the length
	//of any part line and pass it to the next buffer
	//We also check if it is in the last line but past
	//the end of the buffer

	ThisY = cCR * LineHeight;
	
	if (Y > ThisY){

		//This is the same logic as in BufferToClient()
		//See that function for an extended discussion!

		if (cCR){

			i = Size;
            i--;
			for (i=Size-1;i>=0 && Buffer[i] != '\n';i--);

		}

		else
			i = -1;

		i += 1;

		dw = GetTabbedTextExtent(DC,Buffer+i,Size-i,1,&TabSize);

		if (cCR)
			XOffset = 0;

		XOffset += LOWORD(dw);

		if (Y>ThisY+LineHeight || XOffset < X)
			if (!Next)
				return CharsSoFar + Size; //Off the end of the buffer
			else
				return Next->ClientToBuffer(X,Y-ThisY,DC,CharsSoFar+Size,XOffset);

	}

	//If we get here then the position is definately in this buffer
	//Scan forward enough lines to satisfy the Y co-ord.
	//If we need any at all the X offset is no longer
	//relevant as it is only for part lines. This logic is flakey
	//if this is the first buffer and there are no CRs at all so we
    //put in a special little trap for that case.

	NeededCR = Y / LineHeight;

	if (NeededCR)
		XOffset = 0;

	if (!Prev && !cCR)
		XOffset = 0;

	for (i=0;NeededCR>0;i++){
		
		if (i == Size) //this shouldn't happen
			return CharsSoFar + Size;

		if (Buffer[i] == '\n')
			NeededCR--;

	}

	//Now we are in the right line - keep checking the line length
	//until we pass the X co-ord then return the nearest one. This
	//should really be a binary chop search but it will work for now.

	String = Buffer + i;
	X -= XOffset;
	CharsSoFar += i;
	LastX = 0;

	for(i=0;;i++){

		//The user might have clicked past the end of the line

		if (String[i] == '\n')
			return CharsSoFar + i;

		ThisX = LOWORD(GetTabbedTextExtent(DC,String,i,1,&TabSize));

		if (ThisX > X){

			if (X - LastX < ThisX - X)
				i--;

			return CharsSoFar + i;

		}

		LastX = ThisX;

	}

}

int TBFileBuffer::Height(long BufferPos){

	if (BufferPos < Size || !Next)
		return LineHeight;

	return Next->Height(BufferPos - Size);

}
		
long TBFileBuffer::NextBreak(long BufferPos,long BufferSoFar){

long    i;

	if (BufferPos > Size)
		if (!Next)
			return BufferSoFar + Size;
		else
			return Next->NextBreak(BufferPos - Size,BufferSoFar + Size);
			
	//If we get here the the index points to this buffer
	//Now find the next Word Break
	
	for (i = BufferPos;IsCharAlphaNumeric(Buffer[i]) && i<Size;i++);
	
	if (i == Size && Next) //Break is in next buffer
		return Next->NextBreak(0,BufferSoFar + Size);
			
	return BufferSoFar + i;
	 
}

long TBFileBuffer::PrevBreak(long BufferPos,long BufferSoFar){

long    i;

	if (BufferPos > Size)
		if (!Next)
			return BufferSoFar + Size;
		else
			return Next->PrevBreak(BufferPos - Size,BufferSoFar + Size);
			
	//If we get here the the index points to this buffer
	//Now find the previous Break
	
	for (i = BufferPos-1;IsCharAlphaNumeric(Buffer[i]) && i>=0;i--);
	
	if (i < 0) //Break is in prev buffer
		if (Prev)
			return BufferSoFar - Prev->LastChars() + 1;
		else
			return 0;
			
	return BufferSoFar + i + 1;
	 
}

long TBFileBuffer::NextCR(long BufferPos,long BufferSoFar){

long    i;

	if (BufferPos > Size)
		if (!Next)
			return BufferSoFar + Size;
		else
			return Next->NextCR(BufferPos - Size,BufferSoFar + Size);
			
	//If we get here the the index points to this buffer
	//Now find the next CR
	
	for (i = BufferPos;Buffer[i] != '\n' && i<Size;i++);
	
	if (i == Size && Next) //CR is in next buffer
		return Next->NextCR(0,BufferSoFar + Size);
			
	return BufferSoFar + i;
	 
}

long TBFileBuffer::PrevCR(long BufferPos,long BufferSoFar){

long    i;

	if (BufferPos > Size)
		if (!Next)
			return BufferSoFar + Size;
		else
			return Next->PrevCR(BufferPos - Size,BufferSoFar + Size);
			
	//If we get here the the index points to this buffer
	//Now find the previous CR
	
	for (i = BufferPos-1;Buffer[i] != '\n' && i>=0;i--);
	
	if (i < 0) //CR is in prev buffer
		if (Prev)
			return BufferSoFar - Prev->LastChars() + 1;
		else
			return 0;
			
	return BufferSoFar + i + 1;
	 
}

void TBFileBuffer::CopyChar(char FAR *Dest,long BufferPos,long CharsToCopy){

long    i;
long	CopyHere;

	if ((BufferPos > Size) && Next){

		Next->CopyChar(Dest,BufferPos - Size,CharsToCopy);
		return;

	}

	//If we get here the the index points to this buffer

	if (BufferPos + CharsToCopy <= Size)
		CopyHere = CharsToCopy;
    else
		CopyHere = Size - BufferPos;
	
	for (i = 0;i < CopyHere;i++)
		Dest[i] = Buffer[i+BufferPos];
	
	//Did we get them all?

	if ((CopyHere != CharsToCopy) && Next)

    	//No
		 
		Next->CopyChar(Dest+CopyHere,0,CharsToCopy-CopyHere);

}

void TBFileBuffer::DeleteAt(long BufferPos,long CharsToDelete){

long    i;
long	DeleteHere;

	if ((BufferPos > Size) && Next){

		Next->DeleteAt(BufferPos - Size,CharsToDelete);
		return;

	}

	//If we get here the the index points to this buffer

	if (BufferPos + CharsToDelete <= Size)
		DeleteHere = CharsToDelete;
    else
		DeleteHere = Size - BufferPos;
	
	for (i = BufferPos;i < Size-DeleteHere;i++)
		Buffer[i] = Buffer[i+DeleteHere];
	
	Size -= DeleteHere;	
	SetStats();

	//Did we get them all?

	if ((DeleteHere != CharsToDelete) && Next)

    	//No
		 
		Next->DeleteAt(0,CharsToDelete-DeleteHere);

}

void TBFileBuffer::InsertAt(long BufferPos,unsigned char Char){

long    i;

	if ((BufferPos > Size) && Next){

		Next->InsertAt(BufferPos - Size,Char);
		return;

	}

	//If we get here the the index points to this buffer
	
	if (Size == BUFFSIZE){

		Split();
	
		//The split pos may have moved to the new buffer

		if (BufferPos > Size){

			Next->InsertAt(BufferPos - Size,Char);
			return;

		}

    }

	for (i = Size;i > BufferPos;i--)
		Buffer[i] = Buffer[i-1];
	Buffer[i] = Char;
	
	Size++;	
	SetStats();
		 
}

void TBFileBuffer::InsertAt(long BufferPos,char* String){

long    i;
int		ToInsert;

	if ((BufferPos > Size) && Next){

		Next->InsertAt(BufferPos - Size,String);
		return;

	}

	//If we get here the the index points to this buffer
	
    ToInsert = strlen(String);

	if (Size + ToInsert > BUFFSIZE){

		Split();
	
		//The split pos may have moved to the new buffer

		if (BufferPos > Size){

			Next->InsertAt(BufferPos - Size,String);
			return;

		}

    }

	for (i = Size+ToInsert-1;i > BufferPos+ToInsert-1;i--)
		Buffer[i] = Buffer[i-ToInsert];


	memcpy(Buffer+BufferPos,String,ToInsert);

	Size += ToInsert;
	SetStats();
		 
}

void TBFileBuffer::Search(TBSearchStruct& SearchStruct){

long    i,j,SearchEnd;
int		SearchLength;

//This logic works fine but needs some major optimizing.
//It is too slow on big files when nothing is found.

	if (!Prev) //i.e. this is the first buffer in the chain
    	SearchStruct.FoundAt = 0;

	if (SearchStruct.StartPos > Size){

		if (!Next){

			SearchStruct.Found = FALSE;
			return;

		} 

		SearchStruct.StartPos -= Size;
		SearchStruct.FoundAt += Size;
		Next->Search(SearchStruct);
        return;

	}

	//If we get here then we are in the search range

	SearchLength = strlen(SearchStruct.SearchText);
	SearchEnd = Size - SearchLength;

	for (i = SearchStruct.StartPos;i <= SearchEnd;i++){

		if ((SearchStruct.CaseSensitive &&
				!strncmp(SearchStruct.SearchText,Buffer+i,SearchLength))
		|| (!SearchStruct.CaseSensitive &&
				!strnicmp(SearchStruct.SearchText,Buffer+i,SearchLength))){

			SearchStruct.Found = TRUE;
               SearchStruct.FoundAt += i; 
			return;

        }

	}

	//If we get here then the search string is not in this buffer
	//but it could be split across this and the next buffer!

	if (!Next){

	 	//Assuming there is a next buffer, of course
		SearchStruct.Found = FALSE;
		return;

    }
				 
	for (j=1;i < Size;i++,j++){

		if ((SearchStruct.CaseSensitive &&
			(!strncmp(SearchStruct.SearchText,Buffer+i,SearchLength-j) &&
			!strncmp(SearchStruct.SearchText+SearchLength-j,Next->Buffer,j)))
		|| (!SearchStruct.CaseSensitive &&
			(!strnicmp(SearchStruct.SearchText,Buffer+i,SearchLength-j) &&
			!strnicmp(SearchStruct.SearchText+SearchLength-j,Next->Buffer,j)))){

			SearchStruct.Found = TRUE;
			SearchStruct.FoundAt += i; 
			return;

        }

	}

	//If we get this far then it is not in this buffer
	//at all - try the next one

	SearchStruct.StartPos = 0;
	SearchStruct.FoundAt += Size;
	Next->Search(SearchStruct);

}

void TBFileBuffer::Split(){

PTBFileBuffer	NewBuffer;
PTBFileBuffer	OldNext;
const int 		CharsToMove = BUFFSIZE/2;

	OldNext = Next;

	NewBuffer = new TBFileBuffer(Parent,this);

	if (!NewBuffer){

		MessageBox(Parent->HWindow,"Unable to allocate buffer - Chain corrupted - Quit Now","Memory Error",MB_OK | MB_ICONEXCLAMATION);
		return;

	}

	//Fix up the chain
	//NewBuffer->Prev & this->Next set by constructor

	if (OldNext){

		OldNext->Prev = NewBuffer;
		NewBuffer->Next = OldNext;

	}

	memcpy(Next->Buffer,Buffer+(BUFFSIZE-CharsToMove),CharsToMove); //Is this right?
	
	Next->Size = CharsToMove;
	Size = BUFFSIZE - CharsToMove;

	SetStats();
    Next->SetStats();

}

/* Scrolls to an (X,Y) position, after checking boundary conditions.
   Causes a WMPaint message to be sent.  First scrolls the contents
   of the client area, if a portion of the client area will
   remain visible. This is a copy of the equivalent TScroller
   proc with one small change. */

void TBScroller::ScrollTo(long X,long Y){

	long NewXPos,NewYPos;

  	if (Window){

    	NewXPos = max(0, min(X, XRange));
    	NewYPos = max(0, min(Y, YRange));

    	if ((NewXPos != XPos) || (NewYPos != YPos)){

    		/* Scaling is not needed here. If condition is met, ScrollWindow()
       		will succeed since XPage and YPage are ints. If condition is not
       		met, ScrollWindow() is called in EndScroll() as a result of calling
			UpdateWindow(). EndScroll() performs the necessary scaling.*/

			//This is the change

//	      	if (AutoOrg || (abs(YPos - NewYPos) < YPage) && (abs(XPos - NewXPos) < XPage))
//    	    	ScrollWindow(Window->HWindow, (int)((XPos-NewXPos)*XUnit),
//                          (int)((YPos-NewYPos)*YUnit), NULL, NULL);
//			else
				InvalidateRect(Window->HWindow,NULL,/*TRUE*/FALSE);

      		XPos = NewXPos;
	      	YPos = NewYPos;
    	  	UpdateWindow(Window->HWindow);

    	}

  	}

}

TBSearchDialog::TBSearchDialog(PTWindowsObject AParent,int ResourceId,PTBSearchStruct SearchStruct,PTModule AModule):TDialog(AParent,ResourceId,AModule){

  	new TEdit(this,ID_SEARCHTEXT,sizeof(SearchStruct->SearchText),GetModule());
  	new TCheckBox(this,ID_CASESENSITIVE,(PTGroupBox)NULL,GetModule());

  	if (ResourceId == SD_REPLACE){

		new TEdit(this,ID_REPLACETEXT,sizeof (SearchStruct->ReplaceText),GetModule());
		new TCheckBox(this,ID_REPLACEALL,(PTGroupBox)NULL,GetModule());
		new TCheckBox(this,ID_PROMPTONREPLACE,(PTGroupBox)NULL,GetModule());

  	}

	TransferBuffer = SearchStruct;

}

TBSearchStruct::TBSearchStruct(){

	SearchText[0]	= 0;
  	CaseSensitive	= FALSE;
  	ReplaceText[0]	= 0;
  	ReplaceAll		= FALSE;
	PromptOnReplace	= FALSE;

    Found			= FALSE;

}